/*
 * GPIO driver for Artila Matrix-500 based on Atmel AT91RM9200 (Thunder)
 *
 *  Copyright (C) 2005 Artila Electronics,Ltd
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/miscdevice.h>
#include <asm/bitops.h> 
#include <asm/uaccess.h>
#include <linux/init.h>
#include <linux/spinlock.h>
#include <mach/hardware.h>
#include <mach/at91_pio.h>
#include <mach/gpio.h>

#define GPIO_MINOR 		254

#define procfs_name "gpio"
struct proc_dir_entry *matrix_proc_file;

#define PIN  0
#define OUTP 1
#define INP  2
#define LOW  0
#define HIGH 1

#define M504_GPIO_COUNT	5
//5
static int M504_GPIO_MAP[M504_GPIO_COUNT]={
	AT91_PIN_PB30,
	AT91_PIN_PB31,
	AT91_PIN_PC0,
	AT91_PIN_PC1,
	AT91_PIN_PC2,
	
	
};


#define M504_TOTAL_GPIO	(M504_GPIO_COUNT)
/* IOCTL Commands */
#define GPIO_IOCTL_COUNT	0
#define GPIO_IOCTL_DIPSW	1
#define GPIO_IOCTL_GET		2
#define	GPIO_IOCTL_SET		3
#define	GPIO_IOCTL_CLEAR	4
#define GPIO_IOCTL_OUTPUT	5
#define GPIO_IOCTL_INPUT	6
#define GPIO_IOCTL_MODE		7
#define GPIO_IOCTL_INPUT_SET	8
#define GPIO_IOCTL_STATUS	0xa001


static DEFINE_SPINLOCK(gpio_lock);
static int gpio_open_cnt=0;

static int m504_gpio_mode[M504_TOTAL_GPIO];
/* ......................................................................... */

/*
 * GPIO device is opened, and GPIO starts running.
 */
static int m504_gpio_open(struct inode *inode, struct file *file)
{
	spin_lock(&gpio_lock);
	gpio_open_cnt++;
	spin_unlock(&gpio_lock);
	return 0;
}

/*
 * Close the GPIO device.
 * If CONFIG_GPIO_NOWAYOUT is NOT defined then the GPIO is also
 *  disabled.
 */
static int m504_gpio_close(struct inode *inode, struct file *file)
{
	spin_lock(&gpio_lock);
	spin_unlock(&gpio_lock);
	return 0;
}

/*
 * Handle commands from user-space.
 */
static int m504_gpio_ioctl(struct inode *inode, struct file *file,
		unsigned int cmd, unsigned long arg)
{
	void __user *argp = (void __user *)arg;
	int i;
	unsigned long status = 0x00000000;
	
	switch(cmd){
		case GPIO_IOCTL_COUNT:
			*((int __user *) argp) = M504_GPIO_COUNT; 
		return 0;
		case GPIO_IOCTL_DIPSW:
			*((int __user *) argp) = 0; 
		return 0;
		case GPIO_IOCTL_GET:
			i = *((int __user *) argp);
			if(i >= M504_TOTAL_GPIO)
				return -EINVAL;	
		return at91_get_gpio_value(M504_GPIO_MAP[i]);

		case GPIO_IOCTL_SET:
			i = *((int __user *) argp);
			if(i >= M504_GPIO_COUNT  )
				return -EINVAL;	
			if(m504_gpio_mode[i] == 1)
				at91_set_gpio_input(M504_GPIO_MAP[i], 1);
			else
				at91_set_gpio_output(M504_GPIO_MAP[i], 1);	
			
		return 0;

		case GPIO_IOCTL_CLEAR:
			i = *((int __user *) argp);
			if(i >= M504_GPIO_COUNT  )
				return -EINVAL;	
			if(m504_gpio_mode[i] == 1)
				at91_set_gpio_input(M504_GPIO_MAP[i], 0);
			else
				at91_set_gpio_output(M504_GPIO_MAP[i], 0);
		return 0;
		
		case GPIO_IOCTL_OUTPUT:

			i = *((int __user *) argp);
			if(i >= M504_GPIO_COUNT )
				return -EINVAL;
			m504_gpio_mode[i] = 0;
			at91_set_gpio_output(M504_GPIO_MAP[i], 0);
			return 0;
		case GPIO_IOCTL_INPUT:
			i = *((int __user *) argp);
			if(i >= M504_GPIO_COUNT)
				return -EINVAL;
			m504_gpio_mode[i] = 1;
			at91_set_gpio_input(M504_GPIO_MAP[i], 0);
			
			return 0;
		case	GPIO_IOCTL_MODE:
			i = *((int __user *) argp);
			if(i >= M504_GPIO_COUNT)
				return -EINVAL;
			return m504_gpio_mode[i];

		/*case 	GPIO_IOCTL_INPUT_SET:
			i = *((int __user *) argp);
			return at91_set_gpio_input(M504_GPIO_MAP[i][0], 1); 
		*/
		case	0xa001:
			for (i = M504_TOTAL_GPIO ; i >= 0 ; i--) {
				status <<= 1;
				if(at91_get_gpio_value(M504_GPIO_MAP[i]))
					status++;
			}
			*((unsigned long __user *) argp) = status; 
			return 0;		
		default:return -ENOIOCTLCMD; 
	}
	return -ENOIOCTLCMD;
}

static ssize_t matrix500_gpio_read(struct file *file, char *data, size_t len, loff_t *ppos)
{
	int i;
	unsigned long status = 0x00000000;
	int bufs[M504_TOTAL_GPIO+1];
	spin_lock(&gpio_lock);
	for (i = M504_TOTAL_GPIO ; i >= 0 ; i--) {
		//status <<= 1;
		//bufs[]
		if(at91_get_gpio_value(M504_GPIO_MAP[i]))
			//status++;
			bufs[i]=1;
		else
			bufs[i]=0;	
	}
	copy_to_user (data, bufs, sizeof(bufs));
	spin_unlock(&gpio_lock);
	return strlen(bufs);
}
static ssize_t matrix500_gpio_write(struct file *file, const char *data, size_t len, loff_t *ppos)
{
	spin_lock(&gpio_lock);
	spin_unlock(&gpio_lock);
	return len;
}
/* ......................................................................... */

static struct file_operations m504gpio_fops =
{
	.owner		= THIS_MODULE,
	.ioctl		= m504_gpio_ioctl,
	.open		= m504_gpio_open,
	.release	= m504_gpio_close,
	.read		= matrix500_gpio_read,
	.write		= matrix500_gpio_write,
};

static struct miscdevice m504_gpio_miscdev =
{
	.minor		= GPIO_MINOR,
	.name		= "gpio",
	.fops		= &m504gpio_fops,	
};

static int __init m504_gpio_init(void)
{
	int res;
	int i;

	res = misc_register(&m504_gpio_miscdev);
	if (res)
		return res;
	
	for(i = 0 ; i < M504_GPIO_COUNT ; i++) {
		at91_set_gpio_input(M504_GPIO_MAP[i],LOW);
		m504_gpio_mode[i] = 1;
	}

	printk("M504 GPIO Driver Ver 1.00 Loaded.\n");
	return 0;
}

static void __exit m504_gpio_exit(void)
{
	misc_deregister(&m504_gpio_miscdev);
	
}

module_init(m504_gpio_init);
module_exit(m504_gpio_exit);

MODULE_LICENSE("GPL")
MODULE_AUTHOR("Ace Yang")
MODULE_DESCRIPTION("GPIO driver for M504 based on AT91SAM9G20")
