/*
 * drivers/i2c/chips/m41t00.c
 *
 * I2C client/driver for the ST M41T00 Real-Time Clock chip.
 *
 * Author: Mark A. Greer <mgreer@mvista.com>
 *
 * 2005 (c) MontaVista Software, Inc. This file is licensed under
 * the terms of the GNU General Public License version 2. This program
 * is licensed "as is" without any warranty of any kind, whether express
 * or implied.
 */
/*
 * This i2c client/driver wedges between the drivers/char/genrtc.c RTC
 * interface and the SMBus interface of the i2c subsystem.
 * It would be more efficient to use i2c msgs/i2c_transfer directly but, as
 * recommened in .../Documentation/i2c/writing-clients section
 * "Sending and receiving", using SMBus level communication is preferred.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/interrupt.h>
#include <linux/i2c.h>
#include <linux/rtc.h>
#include <linux/bcd.h>
#include <linux/m41t00.h>
//#include <asm/time.h>
#include <asm/rtc.h>

#define	M41T00_DRV_NAME		"m41t00"

static DECLARE_MUTEX(m41t00_mutex);

static struct i2c_driver m41t00_driver;
static struct i2c_client *save_client;

static unsigned short ignore[] = { I2C_CLIENT_END };
static unsigned short normal_addr[] = { M41T00_I2C_ADDR, I2C_CLIENT_END };
//static unsigned short normal_addr[] = { I2C_CLIENT_END, I2C_CLIENT_END };

static struct i2c_client_address_data addr_data = {
	.normal_i2c		= normal_addr,
	.probe			= ignore,
	.ignore			= ignore,
};
#if 1
struct m41t00_chip_info {
	u8	type;
	char	*name;
	u8	read_limit;
	u8	sec;		/* Offsets for chip regs */
	u8	min;
	u8	hour;
	u8	day;
	u8	mon;
	u8	year;
	u8	alarm_mon;
	u8	alarm_hour;
	u8	sqw;
	u8	sqw_freq;
};

static struct m41t00_chip_info m41t00_chip_info_tbl[] = {
	{
		.type		= M41T00_TYPE_M41T00,
		.name		= "m41t00",
		.read_limit	= 5,
		.sec		= 0,
		.min		= 1,
		.hour		= 2,
		.day		= 4,
		.mon		= 5,
		.year		= 6,
	},
	{
		.type		= M41T00_TYPE_M41T81,
		.name		= "m41t81",
		.read_limit	= 1,
		.sec		= 1,
		.min		= 2,
		.hour		= 3,
		.day		= 5,
		.mon		= 6,
		.year		= 7,
		.alarm_mon	= 0xa,
		.alarm_hour	= 0xc,
		.sqw		= 0x13,
	},
	{
		.type		= M41T00_TYPE_M41T85,
		.name		= "m41t85",
		.read_limit	= 1,
		.sec		= 1,
		.min		= 2,
		.hour		= 3,
		.day		= 5,
		.mon		= 6,
		.year		= 7,
		.alarm_mon	= 0xa,
		.alarm_hour	= 0xc,
		.sqw		= 0x13,
	},
};

static struct m41t00_chip_info m41t00_chip;

#endif

#if 1

#define M41T80_REG_SSEC	0
#define M41T80_REG_SEC	1
#define M41T80_REG_MIN	2
#define M41T80_REG_HOUR	3
#define M41T80_REG_WDAY	4
#define M41T80_REG_DAY	5
#define M41T80_REG_MON	6
#define M41T80_REG_YEAR	7
#define M41T80_REG_ALARM_MON	0xa
#define M41T80_REG_ALARM_DAY	0xb
#define M41T80_REG_ALARM_HOUR	0xc
#define M41T80_REG_ALARM_MIN	0xd
#define M41T80_REG_ALARM_SEC	0xe
#define M41T80_REG_FLAGS	0xf
#define M41T80_REG_SQW	0x13

#define M41T80_DATETIME_REG_SIZE	(M41T80_REG_YEAR + 1)
#define M41T80_ALARM_REG_SIZE	\
	(M41T80_REG_ALARM_SEC + 1 - M41T80_REG_ALARM_MON)

#define M41T80_SEC_ST		(1 << 7)	/* ST: Stop Bit */
#define M41T80_ALMON_AFE	(1 << 7)	/* AFE: AF Enable Bit */
#define M41T80_ALMON_SQWE	(1 << 6)	/* SQWE: SQW Enable Bit */
#define M41T80_ALHOUR_HT	(1 << 6)	/* HT: Halt Update Bit */
#define M41T80_FLAGS_AF		(1 << 6)	/* AF: Alarm Flag Bit */
#define M41T80_FLAGS_BATT_LOW	(1 << 4)	/* BL: Battery Low Bit */

#define M41T80_FEATURE_HT	(1 << 0)
#define M41T80_FEATURE_BL	(1 << 1)

static int m41t80_get_datetime(struct i2c_client *client,
			       struct rtc_time *tm)
{
	u8 buf[M41T80_DATETIME_REG_SIZE], dt_addr[1] = { M41T80_REG_SEC };
	struct i2c_msg msgs[] = {
		{
			.addr	= client->addr,
			.flags	= 0,
			.len	= 1,
			.buf	= dt_addr,
		},
		{
			.addr	= client->addr,
			.flags	= I2C_M_RD,
			.len	= M41T80_DATETIME_REG_SIZE - M41T80_REG_SEC,
			.buf	= buf + M41T80_REG_SEC,
		},
	};

	if (i2c_transfer(client->adapter, msgs, 2) < 0) {
		dev_err(&client->dev, "read error\n");
		return -EIO;
	}

	tm->tm_sec = BCD2BIN(buf[M41T80_REG_SEC] & 0x7f);
	tm->tm_min = BCD2BIN(buf[M41T80_REG_MIN] & 0x7f);
	tm->tm_hour = BCD2BIN(buf[M41T80_REG_HOUR] & 0x3f);
	tm->tm_mday = BCD2BIN(buf[M41T80_REG_DAY] & 0x3f);
	tm->tm_wday = buf[M41T80_REG_WDAY] & 0x07;
	tm->tm_mon = BCD2BIN(buf[M41T80_REG_MON] & 0x1f) - 1;

	/* assume 20YY not 19YY, and ignore the Century Bit */
	tm->tm_year = BCD2BIN(buf[M41T80_REG_YEAR]) + 100;
	return 0;
}

/* Sets the given date and time to the real time clock. */
static int m41t80_set_datetime(struct i2c_client *client, struct rtc_time *tm)
{
	u8 wbuf[1 + M41T80_DATETIME_REG_SIZE];
	u8 *buf = &wbuf[1];
	u8 dt_addr[1] = { M41T80_REG_SEC };
	struct i2c_msg msgs_in[] = {
		{
			.addr	= client->addr,
			.flags	= 0,
			.len	= 1,
			.buf	= dt_addr,
		},
		{
			.addr	= client->addr,
			.flags	= I2C_M_RD,
			.len	= M41T80_DATETIME_REG_SIZE - M41T80_REG_SEC,
			.buf	= buf + M41T80_REG_SEC,
		},
	};
	struct i2c_msg msgs[] = {
		{
			.addr	= client->addr,
			.flags	= 0,
			.len	= 1 + M41T80_DATETIME_REG_SIZE,
			.buf	= wbuf,
		 },
	};

	/* Read current reg values into buf[1..7] */
	if (i2c_transfer(client->adapter, msgs_in, 2) < 0) {
		dev_err(&client->dev, "read error\n");
		return -EIO;
	}

	wbuf[0] = 0; /* offset into rtc's regs */
	/* Merge time-data and register flags into buf[0..7] */
	buf[M41T80_REG_SSEC] = 0;
	buf[M41T80_REG_SEC] =
		BIN2BCD(tm->tm_sec) | (buf[M41T80_REG_SEC] & ~0x7f);
	buf[M41T80_REG_MIN] =
		BIN2BCD(tm->tm_min) | (buf[M41T80_REG_MIN] & ~0x7f);
	buf[M41T80_REG_HOUR] =
		BIN2BCD(tm->tm_hour) | (buf[M41T80_REG_HOUR] & ~0x3f) ;
	buf[M41T80_REG_WDAY] =
		(tm->tm_wday & 0x07) | (buf[M41T80_REG_WDAY] & ~0x07);
	buf[M41T80_REG_DAY] =
		BIN2BCD(tm->tm_mday) | (buf[M41T80_REG_DAY] & ~0x3f);
	buf[M41T80_REG_MON] =
		BIN2BCD(tm->tm_mon + 1) | (buf[M41T80_REG_MON] & ~0x1f);
	/* assume 20YY not 19YY */
	buf[M41T80_REG_YEAR] = BIN2BCD(tm->tm_year % 100);

	if (i2c_transfer(client->adapter, msgs, 1) != 1) {
		dev_err(&client->dev, "write error\n");
		return -EIO;
	}
	return 0;
}

int m41t00_command(struct i2c_client *client,unsigned int cmd,
			  void *arg)
{
	
	dev_dbg(&save_client->dev, "%s: cmd=%d\n", __FUNCTION__, cmd);

	switch (cmd) {
	case 0:
		return m41t80_get_datetime(save_client, arg);

	case 1:
		return m41t80_set_datetime(save_client, arg);
	
	default:
		return -EINVAL;
	}
}
EXPORT_SYMBOL_GPL(m41t00_command);



#else
ulong
m41t00_get_rtc_time(void)
{
	s32 sec, min, hour, day, mon, year;
	s32 sec1, min1, hour1, day1, mon1, year1;
	u8 reads = 0;
	u8 buf[8], msgbuf[1] = { 0 }; /* offset into rtc's regs */
	struct i2c_msg msgs[] = {
		{
			.addr	= save_client->addr,
			.flags	= 0,
			.len	= 1,
			.buf	= msgbuf,
		},
		{
			.addr	= save_client->addr,
			.flags	= I2C_M_RD,
			.len	= 8,
			.buf	= buf,
		},
	};

	sec = min = hour = day = mon = year = 0;

	do {
		if (i2c_transfer(save_client->adapter, msgs, 2) < 0)
			goto read_err;

		sec1 = sec;
		min1 = min;
		hour1 = hour;
		day1 = day;
		mon1 = mon;
		year1 = year;

		sec = buf[m41t00_chip.sec] & 0x7f;
		min = buf[m41t00_chip.min] & 0x7f;
		hour = buf[m41t00_chip.hour] & 0x3f;
		day = buf[m41t00_chip.day] & 0x3f;
		mon = buf[m41t00_chip.mon] & 0x1f;
		year = buf[m41t00_chip.year];
	} while ((++reads < m41t00_chip.read_limit) && ((sec != sec1)
			|| (min != min1) || (hour != hour1) || (day != day1)
			|| (mon != mon1) || (year != year1)));
	
	if ((m41t00_chip.read_limit > 1) && ((sec != sec1) || (min != min1)
			|| (hour != hour1) || (day != day1) || (mon != mon1)
			|| (year != year1)))
		goto read_err;

	sec = BCD2BIN(sec);
	min = BCD2BIN(min);
	hour = BCD2BIN(hour);
	day = BCD2BIN(day);
	mon = BCD2BIN(mon);
	year = BCD2BIN(year);
	//printk("year = %d\n",year);
	year += (year >= 70)?1900:2000;
	//printk("year = %d\n",year);
	/*year += 1900;
	if (year < 1970)
		year += 100;
*/
	return mktime(year, mon, day, hour, min, sec);

read_err:
	dev_err(&save_client->dev, "m41t00_get_rtc_time: Read error\n");
	return 0;
}
EXPORT_SYMBOL_GPL(m41t00_get_rtc_time);


static void
m41t00_set(void *arg)
{
	struct rtc_time	tm;
	int nowtime = *(int *)arg;
	s32 sec, min, hour, day, mon, year;
	u8 wbuf[9], *buf = &wbuf[1], msgbuf[1] = { 0 };
	struct i2c_msg msgs[] = {
		{
			.addr	= save_client->addr,
			.flags	= 0,
			.len	= 1,
			.buf	= msgbuf,
		},
		{
			.addr	= save_client->addr,
			.flags	= I2C_M_RD,
			.len	= 8,
			.buf	= buf,
		},
	};
	
	rtc_time_to_tm(nowtime, &tm);
	tm.tm_year = (tm.tm_year - 1900) % 100;

	sec = BIN2BCD(tm.tm_sec);
	min = BIN2BCD(tm.tm_min);
	hour = BIN2BCD(tm.tm_hour);
	day = BIN2BCD(tm.tm_mday);
	mon = BIN2BCD(tm.tm_mon);
	year = BIN2BCD(tm.tm_year);

	/* Read reg values into buf[0..7]/wbuf[1..8] */
	if (i2c_transfer(save_client->adapter, msgs, 2) < 0) {
		dev_err(&save_client->dev, "m41t00_set: Read error\n");
		return;
	}

	wbuf[0] = 0; /* offset into rtc's regs */
	buf[m41t00_chip.sec] = (buf[m41t00_chip.sec] & ~0x7f) | (sec & 0x7f);
	buf[m41t00_chip.min] = (buf[m41t00_chip.min] & ~0x7f) | (min & 0x7f);
	buf[m41t00_chip.hour] = (buf[m41t00_chip.hour] & ~0x3f) | (hour& 0x3f);
	buf[m41t00_chip.day] = (buf[m41t00_chip.day] & ~0x3f) | (day & 0x3f);
	buf[m41t00_chip.mon] = (buf[m41t00_chip.mon] & ~0x1f) | (mon & 0x1f);
	//Ace
	buf[m41t00_chip.year] = (buf[m41t00_chip.year] & ~0xff) | (year & 0xff);

	if (i2c_master_send(save_client, wbuf, 9) < 0)
		dev_err(&save_client->dev, "m41t00_set: Write error\n");
}

static ulong	new_time;

DECLARE_TASKLET_DISABLED(m41t00_tasklet, m41t00_set, (ulong)&new_time);

int
m41t00_set_rtc_time(ulong nowtime)
{
	new_time = nowtime;

	if (in_interrupt())
		tasklet_schedule(&m41t00_tasklet);
	else
		m41t00_set(&new_time);

	return 0;
}
EXPORT_SYMBOL_GPL(m41t00_set_rtc_time);
#endif

#ifdef CONFIG_M41T81_WDT
void wdt_ping(int wdt_margin)
{
	unsigned char i2c_data[2];
	struct i2c_msg msgs1[1] = {
		{
			.addr	= save_client->addr,
			.flags	= 0,
			.len	= 2,
			.buf	= i2c_data,
		},
	};
	i2c_data[0] = 0x09;		/* watchdog register */

	if (wdt_margin > 31)
		i2c_data[1] = (wdt_margin & 0xFC) | 0x83; /* resolution = 4s */
	else
		/*
		 * WDS = 1 (0x80), mulitplier = WD_TIMO, resolution = 1s (0x02)
		 */
		i2c_data[1] = wdt_margin<<2 | 0x82;

	i2c_transfer(save_client->adapter, msgs1, 1);
}
EXPORT_SYMBOL_GPL(wdt_ping);
/**
 *	wdt_disable:
 *
 *	disables watchdog.
 */
void wdt_disable(void)
{
	unsigned char i2c_data[2], i2c_buf[0x10];
	struct i2c_msg msgs0[2] = {
		{
			.addr	= save_client->addr,
			.flags	= 0,
			.len	= 1,
			.buf	= i2c_data,
		},
		{
			.addr	= save_client->addr,
			.flags	= I2C_M_RD,
			.len	= 1,
			.buf	= i2c_buf,
		},
	};
	struct i2c_msg msgs1[1] = {
		{
			.addr	= save_client->addr,
			.flags	= 0,
			.len	= 2,
			.buf	= i2c_data,
		},
	};

	i2c_data[0] = 0x09;
	i2c_transfer(save_client->adapter, msgs0, 2);

	i2c_data[0] = 0x09;
	i2c_data[1] = 0x00;
	i2c_transfer(save_client->adapter, msgs1, 1);
}
EXPORT_SYMBOL_GPL(wdt_disable);
#endif


/*
 *****************************************************************************
 *
 *	Driver Interface
 *
 *****************************************************************************
 */
void init_m41t00_chip_info(int i) {
	memcpy(&m41t00_chip , &m41t00_chip_info_tbl[i] , sizeof(struct m41t00_chip_info));
	//m41t00_chip = &m41t00_chip_info_tbl[1];
	//m41t00_chip.type = M41T00_TYPE_M41T81;
	//printk("name = %s , sqw = %x\n",m41t00_chip.name,m41t00_chip.sqw_freq );//= 0x13;
}

static int
m41t00_probe(struct i2c_adapter *adap, int addr, int kind)
{
	struct i2c_client *client;
	int rc;

	client = kmalloc(sizeof(struct i2c_client), GFP_KERNEL);
	if (!client)
		return -ENOMEM;

	memset(client, 0, sizeof(struct i2c_client));
	strncpy(client->name, M41T00_DRV_NAME, I2C_NAME_SIZE);
	client->flags = I2C_DF_NOTIFY;
	client->addr = addr;
	client->adapter = adap;
	client->driver = &m41t00_driver;

	if ((rc = i2c_attach_client(client)) != 0) {
		kfree(client);
		return rc;
	}
	/*Artila Ace */
	init_m41t00_chip_info(1);
	
	
	if (m41t00_chip.type != M41T00_TYPE_M41T00) {
		/* If asked, disable SQW, set SQW frequency & re-enable */
		if (m41t00_chip.sqw_freq)
			if (((rc = i2c_smbus_read_byte_data(client,
					m41t00_chip.alarm_mon)) < 0)
			 || ((rc = i2c_smbus_write_byte_data(client,
					m41t00_chip.alarm_mon, rc & ~0x40)) <0)
			 || ((rc = i2c_smbus_write_byte_data(client,
					m41t00_chip.sqw,
					m41t00_chip.sqw_freq)) < 0)
			 || ((rc = i2c_smbus_write_byte_data(client,
					m41t00_chip.alarm_mon, rc | 0x40)) <0))
				goto sqw_err;

		/* Make sure HT (Halt Update) bit is cleared */
		if ((rc = i2c_smbus_read_byte_data(client,
				m41t00_chip.alarm_hour)) < 0)
			goto ht_err;

		if (rc & 0x40)
			if ((rc = i2c_smbus_write_byte_data(client,
					m41t00_chip.alarm_hour, rc & ~0x40))<0)
				goto ht_err;
	}

	/* Make sure ST (stop) bit is cleared */
	if ((rc = i2c_smbus_read_byte_data(client, m41t00_chip.sec)) < 0)
		goto st_err;

	if (rc & 0x80)
		if ((rc = i2c_smbus_write_byte_data(client, m41t00_chip.sec,
				rc & ~0x80)) < 0)
			goto st_err;


	save_client = client;
	printk("ST M41T81 Found\n");
	return 0;
st_err:
	dev_err(&client->dev, "m41t00_probe: Can't clear ST bit\n");
	goto attach_err;
ht_err:
	dev_err(&client->dev, "m41t00_probe: Can't clear HT bit\n");
	goto attach_err;
sqw_err:
	dev_err(&client->dev, "m41t00_probe: Can't set SQW Frequency\n");
attach_err:
	kfree(client);
	return rc;
}

static int
m41t00_attach(struct i2c_adapter *adap)
{
	return i2c_probe(adap, &addr_data, m41t00_probe);
}

static int
m41t00_detach(struct i2c_client *client)
{
	int	rc;

	if ((rc = i2c_detach_client(client)) == 0) {
		kfree(client);
		//tasklet_kill(&m41t00_tasklet);
	}
	return rc;
}

static struct i2c_driver m41t00_driver = {
	.owner		= THIS_MODULE,
	.name		= M41T00_DRV_NAME,
	.id		= I2C_DRIVERID_STM41T00,
	.flags		= I2C_DF_NOTIFY,
	.attach_adapter	= m41t00_attach,
	.detach_client	= m41t00_detach,
	.command	= m41t00_command,
};

static int __init
m41t00_init(void)
{
	return i2c_add_driver(&m41t00_driver);
}

static void __exit
m41t00_exit(void)
{
	i2c_del_driver(&m41t00_driver);
	return;
}

module_init(m41t00_init);
module_exit(m41t00_exit);

MODULE_AUTHOR("Mark A. Greer <mgreer@mvista.com>");
MODULE_DESCRIPTION("ST Microelectronics M41T00 RTC I2C Client Driver");
MODULE_LICENSE("GPL");
